/* $Id: parse.y,v 1.8 2000/11/27 20:09:44 malekith Exp $ */
%{
#define YYERROR_VERBOSE
#define YYSTYPE	struct token *
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "token.h"
#define NEW new_token
#define NS(a, b, c)	NEW(a, b, sel_syn, syn_ ## c)
#define NK(a)		NEW(0, 0, sel_kw, a)
extern struct token *root;
void yyerror(const char *s);
int yylex(void);
%}

%token TOK_NUM		/* int or float */
%token TOK_STRING
%token TOK_ID
%token TOK_KW_CLASS
%token TOK_KW_NULL

%token TOK_KW_INT
%token TOK_KW_FLOAT
%token TOK_KW_STRING
%token TOK_KW_PTR
%token TOK_KW_VOID

%token TOK_KW_NEW
%token TOK_KW_DELETE

%token TOK_KW_PROC
%token TOK_KW_FOR
%token TOK_KW_DO
%token TOK_KW_WHILE
%token TOK_KW_ELSE
%token TOK_KW_IF
%token TOK_KW_RETURN

%token TOK_ARROW	/* -> */
%token TOK_RANGE	/* :: */
%token TOK_2_OR
%token TOK_2_AND
%token TOK_2_XOR
%token TOK_2_PLUS	/* ++ */
%token TOK_2_MINUS	/* -- */
%token TOK_PRE_2_MINUS	/* -- */
%token TOK_PRE_2_PLUS	/* ++ */
%token TOK_POST_2_PLUS	/* -- */
%token TOK_POST_2_MINUS	/* -- */
%token TOK_SL	/* << */
%token TOK_SR	/* >> */
%token TOK_NE	/* != */
%token TOK_LE	/* <= */
%token TOK_GE	/* >= */
%token TOK_EQ	/* == */
%token TOK_NEG	/* unary - */
%token TOK_NOTNEG	/* unary + */
%token TOK_EPLUS
%token TOK_EMINUS
%token TOK_ESTAR
%token TOK_ESLASH
%token TOK_EPERC
%token TOK_EXOR
%token TOK_EOR
%token TOK_EAND
%token TOK_ESR
%token TOK_ESL

%left TOK_KW_ELSE
%left TOK_KW_IF
%right ';'
%left ','
%right '=' TOK_EPLUS TOK_EMINUS TOK_ESTAR TOK_ESLASH TOK_EPERC TOK_EXOR TOK_EOR TOK_EAND TOK_ESR TOK_ESL
%left TOK_2_OR
%left TOK_2_XOR
%left TOK_2_AND
%left '|'
%left '^'
%left '&'
%left TOK_EQ TOK_NE
%left '<' '>' TOK_LE TOK_GE
%left TOK_SR TOK_SL
%left '+' '-'
%left '*' '/' '%'
%right TOK_2_PLUS TOK_2_MINUS TOK_NEG TOK_NOTNEG '!' '~'
%left TOK_ARROW
%left TOK_RANGE
%right TOK_KW_NEW TOK_KW_DELETE
/* %expect 1*/

%%

file:
	/* nothing */		{ $$ = 0; }
	| class file		{ root = $$ = NS($1, $2, file); }
	| func file		{ root = $$ = NS($1, $2, file); }
;

class:
	class_name '{' class_content '}' ';'
		{ $$ = NS($1, $3, class); }
;

class_name:
	TOK_KW_CLASS TOK_ID ':' TOK_ID	{ $$ = NS($2, $4, class_name); }
	| TOK_KW_CLASS TOK_ID		{ $$ = NS($2, 0, class_name); }
;

class_content:
	class_vars TOK_KW_PROC class_procs
		{ $$ = NS($1, $3, class_content); }
	| class_vars
		{ $$ = NS($1, 0, class_content); }
;

/* VARIABLES */

class_vars:
	/* nothing */		{ $$ = 0; }
	| class_vars var	{ $$ = NS($1, $2, class_vars); }
;


var:
	type var_names ';'	{ $$ = NS($1, $2, var); }
;

var_names:
	var_names ',' one_var 	{ $$ = NS($1, $3, var_names); }
	| one_var		{ $$ = $1; }
;

one_var:	TOK_ID	{ $$ = $1; };

/*
one_var:
	TOK_ID '=' exp		{ $$ = NS($1, $3, one_var); } 
	| TOK_ID		{ $$ = NS($1, 0, one_var); }
;*/

/* PROCS : function declarations */

class_procs:
	/* nothing */		{ $$ = 0; }
	| class_procs proc	{ $$ = NS($1, $2, class_procs); }
;

proc:
	vtype TOK_ID '(' proc_args ')' ';'
		{ $$ = NS(NS($1, $2, proc_sign), $4, proc); }
	| vtype TOK_ID '(' proc_args ')' '@' TOK_NUM ';'
		{ 
		  $$ = NS(NS($1, $2, proc_sign), 
			  NS($4, $7, syscall_args), 
			  syscall); 
		}
;

proc_args:
	/* nothing */		{ $$ = 0; }
	| proc_arg		{ $$ = $1; }
	| proc_args ',' proc_arg { $$ = NS($1, $3, proc_args); }
;

proc_arg:
	type TOK_ID			{ $$ = NS($1, $2, proc_arg); }
	| type				{ $$ = NS($1, 0, proc_arg); }
;

/* FUNCTIONS : function definitions */
func:
	vtype TOK_ID TOK_RANGE TOK_ID '(' func_args ')' block 
		{ $$ = NS(NS($1, NS($2, $4, func_name), func_info), 
			  NS($6, $8, func_body), func); }
;

func_args:
	/* nothing */		{ $$ = 0; }
	| func_arg		{ $$ = $1; }
	| func_args ',' func_arg { $$ = NS($1, $3, func_args); }
;

func_arg:
	type TOK_ID			{ $$ = NS($1, $2, func_arg); }
;

/* TYPES */
vtype:
	TOK_KW_VOID		{ $$ = NK(TOK_KW_VOID); }
	| type			{ $$ = $1; }
;

type:
	TOK_KW_INT		{ $$ = NK(TOK_KW_INT); }
	| TOK_KW_FLOAT		{ $$ = NK(TOK_KW_FLOAT); }
	| TOK_KW_STRING		{ $$ = NK(TOK_KW_STRING); }
	| TOK_KW_PTR '(' TOK_ID ')'
		{ $$ = NK(TOK_KW_PTR); $$->left = $3; }
;

/* BLOCK */
block:
	'{' '}'			{ $$ = 0; }
	| '{' stmts '}'	{ $$ = $2; }
;

stmts:
	stmt			{ $$ = $1; }
	| stmt stmts		{ $$ = NEW($1, $2, sel_op, ';'); }
;


/* STATEMENT */
stmt: 
	';'			{ $$ = 0; }
	| control		{ $$ = $1; }
	| exp ';'		{ $$ = $1; }
	| comma_exp ';'		{ $$ = $1; }
	| block			{ $$ = $1; }
	| lvars			{ $$ = $1; }
;

lvars:
	type lvar_names ';'	{ $$ = NS($1, $2, lvars); }
;

lvar_names:
	lvar_names ',' one_lvar 	{ $$ = NS($1, $3, lvar_names); }
	| one_lvar		{ $$ = $1; }
;

one_lvar:
	TOK_ID '=' exp		{ $$ = NS($1, $3, one_lvar); } 
	| TOK_ID		{ $$ = NS($1, 0, one_lvar); }
;

alternative:
	stmt TOK_KW_ELSE stmt		{ $$ = NS($1, $3, alternative); }
	| stmt			{ $$ = NS($1, 0, alternative); }
;

for_exp:
	/* nothing */		{ $$ = 0; }
	| comma_exp		{ $$ = $1; }
	| exp			{ $$ = $1; }
;

control:
	TOK_KW_IF '(' exp ')' alternative 
		{ $$ = NS($3, $5, if); }
	| TOK_KW_WHILE '(' exp ')' stmt
		{ $$ = NS($3, $5, while); }
	| TOK_KW_FOR '(' for_exp ';' for_exp ';' for_exp ')' stmt
		{ $$ = NEW($3, NS($5, NEW($9, $7, sel_op, ';'), 
			while), sel_op, ';'); }
	| TOK_KW_DO stmt TOK_KW_WHILE '(' exp ')' ';'
		{ $$ = NS($5, $2, do_while); }
	| TOK_KW_RETURN ';'
		{ $$ = NS(0, 0, return); }
	| TOK_KW_RETURN exp ';'
		{ $$ = NS($2, 0, return); }
;

comma_exp:
	exp ',' comma_exp	{ $$ = NEW($1, $3, sel_op, ','); }
	| exp ',' exp		{ $$ = NEW($1, $3, sel_op, ','); }
;

/* EXPRESSION */
/* rid = range id */
rid:
	TOK_ID				{ $$ = $1; }
	| TOK_ID TOK_RANGE TOK_ID	{ $$ = NS($1, $3, range); }
;

const:
	string_const			{ $$ = $1; }
	| TOK_NUM			{ $$ = $1; }
	| TOK_KW_NULL			{ $$ = NEW(0, 0, sel_ptr, 0); }
	| rid				{ $$ = $1; }
;

/* allow C-style '"string1" "string2"' */
string_const:
	TOK_STRING			{ $$ = $1; }
	| TOK_STRING string_const	{ $$ = NEW($1, $2, sel_op, '+'); }
;

call_args:
	/* nothing */		{ $$ = 0; }
	| exp			{ $$ = NS($1, 0, call_arg); }
	| call_args ',' exp	{ $$ = NS($1, NS($3, 0, call_arg), call_args); }
;

call:
	rid '(' call_args ')'	{ $$ = NS($1, $3, call); }
;

pcall:
	rid TOK_ARROW call		{ $$ = NS($1, $3, pcall); }
;

pvar:
	rid TOK_ARROW rid		{ $$ = NS($1, $3, pvar); }
;

exp:
	const	 		{ $$ = $1; }
	| call			{ $$ = $1; }
	| pcall			{ $$ = $1; }
	| pvar			{ $$ = $1; }
	| '(' exp ')'		{ $$ = $2; }
	| exp '=' exp		{ $$ = NEW($1, $3, sel_op, '='); }
	| exp TOK_ESR exp		{ $$ = NEW($1, $3, sel_op, TOK_ESR); }
	| exp TOK_ESL exp		{ $$ = NEW($1, $3, sel_op, TOK_ESL); }
	| exp TOK_EPLUS exp		{ $$ = NEW($1, $3, sel_op, TOK_EPLUS); }
	| exp TOK_EMINUS exp	{ $$ = NEW($1, $3, sel_op, TOK_EMINUS); }
	| exp TOK_ESTAR exp		{ $$ = NEW($1, $3, sel_op, TOK_ESTAR); }
	| exp TOK_ESLASH exp	{ $$ = NEW($1, $3, sel_op, TOK_ESLASH); }
	| exp TOK_EPERC exp		{ $$ = NEW($1, $3, sel_op, TOK_EPERC); }
	| exp TOK_EXOR exp		{ $$ = NEW($1, $3, sel_op, TOK_EXOR); }
	| exp TOK_EOR exp		{ $$ = NEW($1, $3, sel_op, TOK_EOR); }
	| exp TOK_EAND exp		{ $$ = NEW($1, $3, sel_op, TOK_EAND); }
	| exp TOK_2_OR exp	{ $$ = NEW($1, $3, sel_op, TOK_2_OR); }
	| exp TOK_2_XOR exp	{ $$ = NEW($1, $3, sel_op, TOK_2_XOR); }
	| exp TOK_2_AND exp	{ $$ = NEW($1, $3, sel_op, TOK_2_AND); }
	| exp '|' exp		{ $$ = NEW($1, $3, sel_op, '|'); }
	| exp '^' exp		{ $$ = NEW($1, $3, sel_op, '^'); }
	| exp '&' exp		{ $$ = NEW($1, $3, sel_op, '&'); }
	| exp '+' exp		{ $$ = NEW($1, $3, sel_op, '+'); }
	| exp '-' exp		{ $$ = NEW($1, $3, sel_op, '-'); }
	| exp '*' exp		{ $$ = NEW($1, $3, sel_op, '*'); }
	| exp '/' exp		{ $$ = NEW($1, $3, sel_op, '/'); }
	| exp '%' exp		{ $$ = NEW($1, $3, sel_op, '%'); }
	| '-' exp %prec TOK_NEG	{ $$ = NEW(0, $2, sel_op, TOK_NEG); }
	| '+' exp %prec TOK_NOTNEG	{ $$ = NEW(0, $2, sel_op, TOK_NOTNEG); }
	| '~' exp		{ $$ = NEW(0, $2, sel_op, '~'); }
	| '!' exp		{ $$ = NEW(0, $2, sel_op, '!'); }
	| TOK_2_PLUS exp		{ $$ = NEW(0, $2, sel_op, TOK_PRE_2_PLUS); }
	| TOK_2_MINUS exp		{ $$ = NEW(0, $2, sel_op, TOK_PRE_2_MINUS); }
	| exp TOK_2_PLUS		{ $$ = NEW(0, $1, sel_op, TOK_POST_2_PLUS); }
	| exp TOK_2_MINUS		{ $$ = NEW(0, $1, sel_op, TOK_POST_2_MINUS); }
	| exp TOK_EQ exp		{ $$ = NEW($1, $3, sel_op, TOK_EQ); }
	| exp TOK_NE exp		{ $$ = NEW($1, $3, sel_op, TOK_NE); }
	| exp TOK_LE exp		{ $$ = NEW($1, $3, sel_op, TOK_LE); }
	| exp TOK_GE exp		{ $$ = NEW($1, $3, sel_op, TOK_GE); }
	| exp TOK_SL exp		{ $$ = NEW($1, $3, sel_op, TOK_SL); }
	| exp TOK_SR exp		{ $$ = NEW($1, $3, sel_op, TOK_SR); }
	| exp '<' exp		{ $$ = NEW($1, $3, sel_op, '<'); }
	| exp '>' exp		{ $$ = NEW($1, $3, sel_op, '>'); }
	| TOK_KW_DELETE exp	{ $$ = NS(0, $2, delete); }
	| TOK_KW_NEW TOK_ID	{ $$ = NS(0, $2, new); }
;

%%
